/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: BasicCluePool.java 5226 2009-11-08 18:40:34Z gaellalire $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.pool.impl.enhanced.impl.basic.clue;

import java.util.List;
import java.util.concurrent.Executor;
import java.util.concurrent.locks.Lock;

import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.impl.enhanced.api.IWaitControl;
import org.ow2.util.pool.impl.enhanced.api.IllegalTimeoutException;
import org.ow2.util.pool.impl.enhanced.api.TimeoutPoolException;
import org.ow2.util.pool.impl.enhanced.api.WaiterInterruptedException;
import org.ow2.util.pool.impl.enhanced.api.basic.IPoolItemFactory;
import org.ow2.util.pool.impl.enhanced.api.basic.IPoolItemRemoveListener;
import org.ow2.util.pool.impl.enhanced.api.basic.PoolFactoryBroken;
import org.ow2.util.pool.impl.enhanced.api.basic.accessmanager.IAccessManager;
import org.ow2.util.pool.impl.enhanced.api.basic.clue.IBasicCluePool;
import org.ow2.util.pool.impl.enhanced.api.basic.clue.accessmanager.IClueAccessManager;
import org.ow2.util.pool.impl.enhanced.impl.PoolError;
import org.ow2.util.pool.impl.enhanced.impl.basic.BasicPool;
import org.ow2.util.pool.impl.enhanced.internal.lock.api.ISignalClearableCondition;

/**
 * @author Gael Lalire
 * @param <E> pool item type
 * @param <C> clue type type of clue
 */
public class BasicCluePool<E, C> extends BasicPool<E> implements IBasicCluePool<E, C> {

    private static final Log LOG = LogFactory.getLog(BasicCluePool.class);

    private List<E> availablePoolItemList;

    private List<E> unmodifiableAvailablePoolItemList;

    private IClueAccessManager<? super E, ? super C> clueAccessManager;

    private Lock lock;

    private ISignalClearableCondition signalClearableCondition;

    public BasicCluePool(final IPoolItemFactory<? extends E> poolItemFactory, final int initPoolSize,
            final IClueAccessManager<? super E, ? super C> clueAccessManager, final Executor createItemExecutor,
            final IPoolItemRemoveListener<? super E> poolItemRemoveCB, final Executor removeItemExecutor) {
        super(poolItemFactory, initPoolSize, clueAccessManager, createItemExecutor, poolItemRemoveCB, removeItemExecutor);
        this.clueAccessManager = clueAccessManager;
        availablePoolItemList = getAvailablePoolItemList();
        unmodifiableAvailablePoolItemList = getUnmodifiableAvailablePoolItemList();
        lock = getLock();
        signalClearableCondition = getSignalClearableCondition();
    }

    /**
     * Get behavior is not the same.
     * @throws WaiterInterruptedException
     */
    @Override
    public E get(final IWaitControl timeout) throws TimeoutPoolException, IllegalTimeoutException, WaiterInterruptedException,
            PoolFactoryBroken, InterruptedException {
        return get(null, timeout);
    }

    public E get(final C clue) throws WaiterInterruptedException, PoolFactoryBroken, InterruptedException {
        try {
            return get(clue, INFINITE_WAIT_CONTROL);
        } catch (TimeoutPoolException e) {
            throw new PoolError("get(clue, INFINITE_TIMEOUT) should not return a TimeoutPoolException");
        } catch (IllegalTimeoutException e) {
            throw new PoolError("get(clue, INFINITE_TIMEOUT) should not return a IllegalTimeoutException");
        }
    }

    protected int getOneItem(final C clue) {
        int position = 0;
        try {
            position = clueAccessManager.choosePoolItemToGet(unmodifiableAvailablePoolItemList, clue);
            if (position != IAccessManager.DELAY_GET && (position < 0 || position >= unmodifiableAvailablePoolItemList.size())) {
                LOG.error("Access manager returns {0} which is an invalid value", position);
                position = 0;
            }
        } catch (RuntimeException e) {
            LOG.error("Access manager exception", e);
        }
        return position;
    }

    public E get(final C clue, final IWaitControl timeout) throws TimeoutPoolException, IllegalTimeoutException,
            WaiterInterruptedException, PoolFactoryBroken, InterruptedException {
        E poolItem;
        boolean delayed = false;
        lock.lock();
        try {
            int size = availablePoolItemList.size();
            int poolItemToGetChoosed = IAccessManager.DELAY_GET;
            if (size != 0) {
                poolItemToGetChoosed = getOneItem(clue);
                if (poolItemToGetChoosed == IAccessManager.DELAY_GET) {
                    setDelayedCount(getDelayedCount() + 1);
                    delayed = true;
                }
            } else {
                switch (getFactoryState()) {
                case BROKEN:
                    throw new PoolFactoryBroken(null);
                case TEMPORARY_BROKEN:
                    throw new PoolFactoryBroken(getTemporaryBrokenFactoryEndTime());
                case WORKING:
                    break;
                default:
                    throw new PoolError();
                }
            }

            if (poolItemToGetChoosed == IAccessManager.DELAY_GET) {
                if (!timeout.canContinueToWait()) {
                    throw new TimeoutPoolException();
                }
                while (poolItemToGetChoosed == IAccessManager.DELAY_GET) {
                    if (!timeout.waitOnCondition(signalClearableCondition)) {
                        throw new TimeoutPoolException();
                    }
                    LOG.debug("wait notified or expired");
                    size = availablePoolItemList.size();
                    if (size != 0) {
                        poolItemToGetChoosed = getOneItem(clue);
                        if (!delayed && poolItemToGetChoosed == IAccessManager.DELAY_GET) {
                            setDelayedCount(getDelayedCount() + 1);
                            delayed = true;
                        }
                    } else {
                        if (delayed) {
                            // presume not delayed if no item available
                            setDelayedCount(getDelayedCount() - 1);
                            delayed = false;
                        }
                        switch (getFactoryState()) {
                        case BROKEN:
                            throw new PoolFactoryBroken(null);
                        case TEMPORARY_BROKEN:
                            throw new PoolFactoryBroken(getTemporaryBrokenFactoryEndTime());
                        case WORKING:
                            break;
                        default:
                            throw new PoolError();
                        }
                    }
                }
            }
            int usedPoolItem = getUsedPoolItem();
            usedPoolItem++;
            setUsedPoolItem(usedPoolItem);
            poolItem = availablePoolItemList.remove(poolItemToGetChoosed);
            if (LOG.isDebugEnabled()) {
                LOG.debug("pool {0}/{1} (used/total)", usedPoolItem, availablePoolItemList.size() + usedPoolItem);
            }
        } finally {
            if (delayed) {
                setDelayedCount(getDelayedCount() - 1);
            }
            lock.unlock();
        }
        return poolItem;
    }

}
